/** This file is part of Reactor.
 *  Copyright (c) 2020-2025 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */

export const version = 25016;

import api from '/client/ClientAPI.js';

import Reaction from '/client/Reaction.js';
import Container from '/client/Container.js';
import Data from '/client/Data.js';

import { idSelector, showSysModal, asyncForEach } from './reactor-ui-common.js';

import { _T } from './i18n.js';

import ReactionEditor from "./reaction-editor.js";

import Tab from "./tab.js";

export default (function($) {

    const tabId = "tab-reactions";
    var tabInstance = false;

    class ReactionsTab extends Tab {
        constructor( $parent ) {
            super( tabId, $parent );

            this.editor = null;
        }

        async toggleDetail( event ) {
            const $row = $( event.currentTarget ).closest( 'div.re-reaction' );
            const id = $row.attr( 'id' );
            const $el = $( '#detail-' + id + '.collapse' );
            if ( ! $el.hasClass( 'loaded' ) ) {
                const reaction = await Reaction.getInstance( id );
                if ( reaction ) {
                    await reaction.refresh();
                    const $body = $( '.card-body', $el ).empty();
                    if ( 0 === (reaction.actions || []).length ) {
                        $body.text( _T('This reaction is empty.') );
                    } else {
                        const $list = $( '<ol></ol>' ).appendTo( $body );
                        await asyncForEach( reaction.actions || [], async ( action ) => {
                            try {
                                let str = await ReactionEditor.makeActionDescription( action );
                                $( '<li></li>' ).text( str ).appendTo( $list );
                            } catch( e ) {
                                console.log( e );
                                $( '<li></li>' ).text( "ERROR" ).appendTo( $list );
                            }
                        });
                    }
                    const $rr = $( '<div></div>' )
                        .text( _T('Reaction ID: {0}', "" ) )
                        .appendTo( $body );
                    $( '<span></span>' ).addClass( "user-select-all" )
                        .text( reaction.id )
                        .appendTo( $rr );
                    $el.addClass( 'loaded' );
                } else {
                    return;
                }
            }
            if ( $el.hasClass( 'show' ) ) {
                $el.slideUp( 250 ).removeClass( 'show' );
                $( '.re-reaction-name > i', $row ).removeClass( 'bi-chevron-down' ).addClass( 'bi-chevron-right' );
            } else {
                $el.slideDown( 250 ).addClass( 'show' );
                $( '.re-reaction-name > i', $row ).removeClass( 'bi-chevron-right' ).addClass( 'bi-chevron-down' );
            }
            return false;
        };

        async editReaction( reaction, isNew ) {
            const self = this;
            const $tab = this.getTabElement();
            $tab.empty();
            $tab.data( 'mode', 'edit' ).attr( 'data-mode', 'edit' );

            let newReaction = !!isNew;

            const $ct = $('<div class="container-fluid re-editor-controls"></div>' ).appendTo( $tab );
            const $row = $( '<div class="row"></div>' ).appendTo( $ct );
            $( '<div class="col"><h1></h1></div>' ).appendTo( $row );
            $( `<div class="col text-end mt-3">
      <button class="btn btn-sm btn-success saveconf">${_T(['#edit-button-save','Save'])}</button>
      <button class="btn btn-sm btn-danger revertconf">${_T(['#edit-button-exit','Exit'])}</button>
    </div>` ).appendTo( $row );
            $( 'btn', $row ).prop( 'disabled', true );
            $( 'h1', $row ).text( reaction.name || reaction.id ).on( 'click', function( ev ) {
                const $el = $( ev.currentTarget );
                const $form = $( '<div class="form-group"><input id="newname" class="form-control form-control-sm"></div>' );
                $( 'input', $form ).val( $el.text() );
                showSysModal({
                    title: _T(['#reaction-rename-title','Rename Reaction']),
                    body: $form,
                    buttons: [{
                        label: _T(['#reaction-rename-button-cancel','Cancel']),
                        class: "btn-primary",
                        close: true
                    },{
                        label: _T(['#reaction-rename-button-rename','Rename']),
                        class: "btn-success",
                        event: "rename"
                    }
                ]
                }).then( data => {
                    if ( "rename" === data ) {
                        const newname = $( '#sysdialog #newname' ).val() || "";
                        if ( "" !== newname && newname !== reaction.name ) {
                            self.editor.rename( newname );
                            $el.text( newname );
                        }
                    }
                });
            });
            $( 'button.saveconf', $ct ).on( 'click', async function( ev ) {
                const $el = $( ev.currentTarget );
                if ( ! self.editor.canSave() ) {
                    showSysModal( {
                        title: 'Uncorrected errors remain!',
                        body: 'Please correct the indicated errors before saving.'
                    } );
                    return;
                }
                $( 'button.saveconf', $tab ).prop( 'disabled', true );
                const reaction = self.editor.data;
                console.log("Saving", reaction.id);
                const robj = await Reaction.getInstance( reaction.id );
                const dobj = robj.getDataObject();
                reaction.serial = ( reaction.serial || 0 ) + 1;
                reaction.mdt = Date.now();
                dobj.setValue( reaction );
                dobj.save().then( function() {
                    self.editor.notifySaved();
                    newReaction = false;
                    if ( $el.hasClass( 'saveexit' ) ) {
                        self.closeEditor();
                    }
                }).catch( function( err ) {
                    console.log("save failed",err);
                    showSysModal( {
                        title: _T('#dialog-error-title'),
                        body: _T('The save failed; try again. {0}')
                    } );
                    self.editor.updateSaveControls();
                });
            });
            $( 'button.revertconf', $ct ).on( 'click', async function( /* ev */ ) {
                const reaction = self.editor.data;
                const robj = await Reaction.getInstance( reaction.id );
                if ( self.editor.isModified() ) {
                    showSysModal( {
                        title: _T(['#reactionedit-unsaved-title','Reaction Modified (Unsaved Changes)']),
                        body: _T(['#reactionedit-unsaved-prompt','You have unsaved changes to this reaction. Really discard changes and exit?']),
                        buttons: [{
                            label: _T(['#reactionedit-unsaved-button-discard','Discard Changes']),
                            event: "discard",
                            class: "btn-danger"
                        },{
                            label: _T(['#reactionedit-unsaved-button-cancel','Cancel']),
                            close: true,
                            class: "btn-success"
                        }]
                    }).then( async data => {
                        if ( "discard" === data ) {
                            self.closeEditor();  // close editor first for unsubscribe/disconnects
                            if ( newReaction ) {
                                await robj.delete();
                            }
                        }
                    });
                } else {
                    self.closeEditor();  // close editor first for unsubscribe/disconnects
                    if ( newReaction ) {
                        await robj.delete();
                    }
                }
            });

            this.editor = new ReactionEditor( reaction, $tab, { fixedName: reaction.name || reaction.id } );
            this.editor.editor().on( 'update-controls', function( ev, ui ) {
                /* During editing, various changes make the reaction saveable, or not. */
                console.log("got update-controls");
                $( 'button.saveconf', $tab ).prop( 'disabled', ui.errors || !ui.modified );
                $( 'button.revertconf', $tab ).prop( 'disabled', false );
            }).on( 'modified', function( ev, ui ) {
                /* Sent when the cached reaction is modified (no errors) */
                console.log("got modified");
                $( 'button.saveconf', $tab ).prop( 'disabled', ui.errors || !ui.modified );
                $( 'button.revertconf', $tab ).prop( 'disabled', false );
            });

            this.editor.edit();
            if ( newReaction ) {
                this.editor.signalModified();
            }
        }

        closeEditor() {
            const $tab = this.getTabElement();
            if ( this.editor ) {
                this.editor.close();
                this.editor = null;
                //debugger;
            }
            $tab.empty();
            $tab.removeData( 'mode' ).attr( 'data-mode', null );
            this.getTabElement().trigger( 'activate' );
        }

        setEnabledToggle( $btn, state ) {
            if ( state ) {
                $( 'i', $btn ).removeClass( 'bi-toggle-off' ).addClass( 'bi-toggle-on' );
                $btn.removeClass( 'text-danger' ).addClass( 'text-success' )
                    .attr( 'title', _T("Enabled; click to disable") );
            } else {
                $( 'i', $btn ).removeClass( 'bi-toggle-on' ).addClass( 'bi-toggle-off' );
                $btn.removeClass( 'text-success' ).addClass( 'text-danger' )
                    .attr( 'title', _T("Disabled; click to enable") );
            }
        }

        async updateReactionRow( reaction ) {
            const $row = $( `div.re-reaction#${idSelector( reaction.id )}` );
            // console.log("Row for",reaction.id,"is",$row);

            /*
            const dobj = await Data.getInstance( Container.getInstance( 'states' ), 'reaction_history' );
            if ( dobj && Array.isArray( dobj.value?.history ) ) {
                let el = dobj.value.history.findLast( el => el.reaction_id === reaction.id );
                if ( el ) {
                    $( 'div.re-reaction-lastrun', $row ).text( new Date( el.start_time ).toLocaleString() );
                }
            }
            */

            $( 'span.re-reaction-name', $row ).text( reaction.name || reaction.id )
                .attr( 'title', String( reaction.name ) + " ID " + reaction.id );

            const $el = $( 'button[data-action="disable"]', $row );
            this.setEnabledToggle( $el, !reaction.disabled );
        };

        async handleReactionEvent( event ) {
            //console.log( 'handleReactionEvent', event );
            if ( 'reaction-changed' === event.type ) {
                this.updateReactionRow( event.data );
                this.handleReactionListFilterChange();
            } else if ( 'reaction-deleted' === event.type ) {
                const $row = $( `div.re-reaction#${idSelector( event.data.id )}` );
                $row.remove();
            }
        };

        getReactionRow( reaction ) {
            const $row = $( '<div class="row re-listport-item re-reaction"></div>' ).attr( 'id', reaction.id );
            let $c = $( '<div class="col pt-1"></div>' )
                .appendTo( $row );
            $( `<button class="btn bi-btn bi-btn-white me-2 text-success" title="${_T(['#reaction-action-run','Run Now'])}" data-action="run"><i class="bi bi-play"></i></button>` )
                .appendTo( $c );
            const $e = $( '<span class="re-reaction-name"></span>' ).text( reaction.name ).appendTo( $c );
            $( '<i class="bi bi-chevron-right ms-1"></i>' ).appendTo( $e );

            /* See also updateReactionRow()
            $c = $( '<div class="col text-end re-reaction-lastrun"></div>' )
                .text( "-" )
                .appendTo( $row );
            */

            $c = $( '<div class="col text-end re-reaction-controls"></div>' )
                .appendTo( $row );
            $( '<button class="btn bi-btn bi-btn-white ms-2 text-success" data-action="disable" title="Enable/Disable Reaction"><i class="bi bi-toggle-on"></i></button>' )
                .appendTo( $c );
            $( `<button class="btn bi-btn bi-btn-white ms-2 text-primary" data-action="edit" title="${_T(['#reaction-action-edit','Edit this Reaction'])}"><i class="bi bi-pencil"></i></button>` )
                .appendTo( $c );
            $( `<button class="btn bi-btn bi-btn-white ms-2 text-secondary" data-action="copy" title="${_T(['#reaction-action-copy','Copy Reaction'])}"><i class="bi bi-share-fill"></i></button>` )
                .appendTo( $c );
            $( `<button class="btn bi-btn bi-btn-white ms-4 text-danger" data-action="delete" title="${_T(['#reaction-action-delete','Delete this Reaction'])}"><i class="bi bi-x"></i></button>` )
                .appendTo( $c );

            /* Detail card (collapsible) */
            $( '<div class="w-100"></div>').appendTo( $row );
            $( '<div class="col-xs-12 col-sm-12 collapse"><div class="card card-body">Jeffrey Epstein didn\'t kill himself.</div></div>' )
                .attr( 'id', 'detail-' + reaction.id )
                .appendTo( $row );

            $( 'span.re-reaction-name', $row ).on( 'click', this.toggleDetail.bind( this ) );

            $( 'button[data-action="run"]', $row ).on( 'click', async ( event ) => {
                const $el = $( event.currentTarget );
                const $row = $el.closest( '.re-reaction' );
                const rid = $row.attr( 'id' );
                const reaction = await Reaction.getInstance( rid );
                console.log("Reaction launching",rid);
                reaction.launch();
            });

            const self = this;

            $( 'button[data-action="disable"]', $row ).on( 'click', async function( event ) {
                const $el = $( event.currentTarget );
                const rid = $el.closest( '.re-reaction' ).attr( 'id' );
                const robj = await Reaction.getInstance( rid );
                /* ??? FIXME -- should update based on change notification for underlying object */
                self.setEnabledToggle( $el, robj.disabled );
                robj.enable( robj.disabled );
                return false;
            });

            $( 'button[data-action="copy"]', $row ).on( 'click', function( event ) {
                const $el = $( event.currentTarget );
                const $row = $el.closest( '.re-reaction' );
                const rid = $row.attr( 'id' );
                /* ??? FIXME -- should not be working on dobjs directly */
                Reaction.getInstance( rid ).then( src => {
                    Reaction.create().then( dest => {
                        const dobj = dest.getDataObject();
                        dobj.value.name = src.name + ' Copy';
                        dobj.value.serial = 1;
                        dobj.value.actions = [ ...src.actions ];
                        dobj.save().then( () => {
                            self.getReactionRow( dest ).insertAfter( $row );
                            self.subscribe( dest, self.handleReactionEvent.bind( self ) );
                        });
                    });
                });
                return false;
            });

            $( 'button[data-action="delete"]', $row ).on( 'click', async function( event ) {
                const $el = $( event.currentTarget );
                const $row = $el.closest( '.re-reaction' );
                const rid = $row.attr( 'id' );
                const robj = await Reaction.getInstance( rid );
                showSysModal( {
                    title: _T(['#reaction-delete-title','Confirm Deletion']),
                    body: _T(['#reaction-delete-prompt','Really delete {0:q}? There is no "undo"'], robj.name),
                    buttons: [
                        {
                            label: _T(['#reaction-delete-button-delete','Delete']),
                            class: "btn-danger",
                            event: "delete"
                        },
                        {
                            label: _T(['#reaction-delete-button-cancel','Cancel']),
                            class: "btn-success"
                        }
                    ]
                } ).then( async function( data ) {
                    if ( "delete" === data ) {
                        robj.delete().then( () => {
                            /* Remove the displayed row */
                            $row.remove();
                        });
                    }
                    return data;
                } );
                return false;
            } );

            $( 'button[data-action="edit"]', $row ).on( 'click', async function( ev ) {
                const rid = $( ev.target ).closest( 'div.re-reaction' ).attr( 'id' );
                const reaction = await Reaction.getInstance( rid );
                if ( ! reaction ) {
                    console.log("Reaction " + String(rid) + " not found");
                    return;
                }
                await reaction.refresh();  // make sure we're looking at latest and greatest
                self.editReaction( reaction.getDataObject().value );
            });

            return $row;
        };

        async handleReactionListFilterChange( /* event */ ) {
            const $field = $( '.re-listport-head input#filter-name' );
            const prefix = ( $field.val() || "" ).toLowerCase();
            $( '.re-listport-body .re-reaction' ).each( (ix,el) => {
                const $el = $( el );
                const reName = $( 'span.re-reaction-name', $el ).text().toLowerCase();
                $el.toggle( reName.indexOf( prefix ) >= 0 );
            });
            localStorage.setItem( 'reactionlist_filter_name', prefix );
        };

        async activate( event ) {
            const self = this;
            console.log("Reactions tab activated,", event);
            const $tab = this.getTabElement();
            /* If editing, don't paint this, it's just an event that bubbled up to us */
            if ( this.editor || "edit" === $tab.data( 'mode' ) ) {
                console.log("Stopping event in edit mode", this.editor);
                return false;
            }
            $tab.empty();
            $tab.data( 'mode', 'list' ).attr( 'data-mode', 'list' );
            const $ct = $( '<div class="re-listport container-fluid ps-0"></div>' ).appendTo( $tab );

            const $section = $( `<div class="re-listport-head container-fluid ps-0">
      <div class="row align-middle">
        <div class="col align-top">
          <span class="re-listport-title">${_T(['#reactions-page-title','Global Reactions'])}</span>
        </div>
        <div class="col-auto float-end mt-2">
          <button id="newreaction" class="btn btn-sm btn-success">${_T(['#reactions-button-create','Create Reaction'])}</button>
        </div>
      </div>
    </div>` )
                .appendTo( $ct );

            const handler = this.handleReactionListFilterChange.bind( this );
            let $row = $( '<div class="d-flex flex-row mb-2"></div>' ).appendTo( $section ); /* not listport-item */
            let $col = $( '<div class=""></div>' ).appendTo( $row );
            $( '<input type="search" id="filter-name" class="form-control form-control-sm" autocomplete="off">' )
                .val( localStorage.getItem( 'reactionlist_filter_name' ) || "" )
                .attr( 'placeholder', _T( 'Name Contains:' ) )
                .on( 'keyup', handler )
                .on( 'change', handler )
                .on( 'search', handler )
                .appendTo( $col );

            const $body = $( '<div class="re-listport-body container-fluid ps-0"></div>' )
                .appendTo( $ct );

            const reactions = await Reaction.getGlobalReactions();
            reactions.sort( function( a, b ) {
                return (a.name || a.id).localeCompare( b.name || b.id, undefined, { sensitivity: 'base' } );
            });
            const nrs = reactions.length;
            /* ??? better to use ul/li ? */
            for ( let k=0; k<nrs; ++k ) {
                const reaction = reactions[k];
                $row = this.getReactionRow( reaction );
                $row.appendTo( $body );
                this.updateReactionRow( reaction );
                this.subscribe( reaction, this.handleReactionEvent.bind( this ) );
            }

            $( 'button#newreaction' ).on( 'click', async function() {
                const reaction = await Reaction.create();
                self.editReaction( reaction.getDataObject().value, true );
            });

            this.handleReactionListFilterChange( {} );

            const dobj = await Data.getInstance( Container.getInstance( 'states' ), 'reaction_queue' );
            if ( dobj ) {
                this.subscribe( dobj, ( ev ) => {
                    const dobj = ev.data;
                    //console.log("reaction queue change", dobj.value);
                    self.updateRunningIndicators( dobj );
                });
                this.updateRunningIndicators( dobj );
            }

            return false;
        };

        updateRunningIndicators( dobj ) {
            // Get list of reactions previously running.
            const $tab = this.getTabElement();
            const olds = new Set();
            $( 'span.re-reaction-running', $tab ).closest( 'div.re-reaction' ).each( (ix,el) => {
                olds.add( $( el ).attr( 'id' ) );
            });
            for ( let entry of dobj.value ) {
                if ( olds.has( entry.id ) ) {
                    olds.delete( entry.id );  // still running
                } else {
                    const $row = $( `div.re-reaction#${idSelector(entry.id)}`, $tab );
                    const $p = $( '<span class="re-reaction-running ms-2 spinner-border spinner-border-sm text-success" role="status"></span>' );
                    $( `span.re-reaction-name`, $row ).parent().append($p);
                    $row.stop().clearQueue().css( 'background-color', 'var(--bs-green)' );
                    $row.animate( { backgroundColor: 'transparent' }, 2000);
                }
            }
            // Remove spinners for reactions no longer running.
            for ( let id of olds.values() ) {
                const $row = $( `div.re-reaction#${idSelector(id)}`, $tab );
                $( 'span.re-reaction-running', $row ).remove();
                $row.stop().clearQueue().css( 'background-color', 'var(--bs-green)' );
                $row.animate( { backgroundColor: 'transparent' }, 2000);
            }
        }

        canSuspend() {
            if ( this.editor && this.editor.isModified() ) {
                console.log( "Stopping event in edit mode" );
                return false;
            }
            return true;
        }

        suspending() {
            this.getTabElement().removeData( 'mode' ).attr( 'data-mode', null );
        };
    };

    return {
        "init": function( $main ) {
            if ( ! tabInstance ) {
                tabInstance = new ReactionsTab( $main );
            }
        },
        "tab": () => tabInstance.getTabElement()
    };
})(jQuery);
